/* ----------------------------------------------------------------------------
 * Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
 * ON Semiconductor), All Rights Reserved
 *
 * This code is the property of ON Semiconductor and may not be redistributed
 * in any form without prior written permission from ON Semiconductor.
 * The terms of use and warranty for this code are covered by contractual
 * agreements between ON Semiconductor and the licensee.
 *
 * This is Reusable Code.
 *
 * ------------------------------------------------------------------------- */

#include <string.h>

#include "audioProcessor.h"
#include "logger.h"
#include "scratch.h"
#include "sharedBuffers.h"

#include "codecs/baseDSP/baseDSPCodec.h"
#include "codecs/baseDSP/baseDSPCodecInternal.h"

/* The audio processor has a work flow when handling its buffers, each buffer
 * must transition between the following states in sequence
 */
typedef enum
{
    EMPTY,
    FULL,
    PROCESSING,
    READY
} BufferState;

/* The audio processor must be configured prior to use, during this time it transitions
 * through these states.
 */
typedef enum
{
    NOT_CONFIGURED, CONFIGURING, IS_CONFIGURED
} ProcessorStatus;

/* the coder state is used to manage the internal operations */
typedef struct _CoderState
{
    ProcessorStatus isConfigured;
    pIOProvider ioProvider;
    BufferState status[2];
    uint32_t inLength[2];
    OperationParameters parameters[2];
    uint32_t inBuffer;
    uint32_t outBuffer;
} CoderState;

static CoderState cs;

/* The audio processor can be used in a polling or interrupt driven mode. */
#ifdef USE_LPDSP32_POLL

static void processingComplete(void);

static void readyToProcess(void);

static int pollPending;

/* ----------------------------------------------------------------------------
 * Function     : setupPoll
 * ----------------------------------------------------------------------------
 * Description  : Sets up a polling operation, this adds a delay to ensure
 *                the remote processor can start handling the request
 * Inputs       : None
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void setupPoll(void)
{
    for (int i = 0; i < 10; i++) logInfo(0x000);
    pollPending = 1;
}

/* ----------------------------------------------------------------------------
 * Function     : pollHandler
 * ----------------------------------------------------------------------------
 * Description  : If we are pending an operation then poll to see if
 *                complete, when it is complete invoke the necessary handler
 * Inputs       : codec : a handle to a codec that should be interrogated
 *                for status
 * Outputs      : returns 1 if still pending completion, 0 if the operation
 *                has completed
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static int pollHandler(CODEC codec)
{
    if (pollPending)
    {
        pBaseDSPCodec cdc = getDSPCodec(codec);

        if ((cdc->dspStatus & 0xF) == 4)
        {
            logEnd();
            if (cs.isConfigured == IS_CONFIGURED)
            {
                processingComplete();
            }
            else
            {
                readyToProcess();
            }
            pollPending = 0;
        }
    }

    return 1;
}

#endif /* ifdef USE_LPDSP32_POLL */

/* ----------------------------------------------------------------------------
 * Function     : loadData
 * ----------------------------------------------------------------------------
 * Description  : fetches data from our audio sample provider and stores it
 *                in the current buffer, it also caches the operational
 *                parameters specified for this set of data. Sets the buffer
 *                status to FULL when complete
 * Inputs       : None
 * Outputs      : Always returns 1
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static uint32_t loadData(void)
{
    /* load the currently selected input buffer with data */
    cs.inLength[cs.inBuffer] = ioRead(cs.ioProvider, Buffer.input[cs.inBuffer],
                                      CODEC_INPUT_SIZE);

    /* copy the operational parameters we are currently using */
    memcpy(&cs.parameters[cs.inBuffer], ioParameters(cs.ioProvider),
           sizeof(OperationParameters));

    /* set the buffer status to full and switch to the next buffer */
    cs.status[cs.inBuffer] = FULL;
    cs.inBuffer = 1 - cs.inBuffer;

    return 1;
}

/* ----------------------------------------------------------------------------
 * Function     : configure
 * ----------------------------------------------------------------------------
 * Description  : configures a codec for first use, this is called at the
 *                start of using the audio processor
 * Inputs       : codec : an opaque handle to a codec object which should be
 *                configured
 * Outputs      : Always returns 1
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static uint32_t configure(CODEC codec)
{
    OperationParameters temp;
    memcpy(&temp, &cs.parameters[cs.outBuffer], sizeof(OperationParameters));
    temp.action = CONFIGURE;
    codecSetParameters(codec, &temp);
    cs.isConfigured = CONFIGURING;
    codecConfigure(codec);

#ifdef USE_LPDSP32_POLL
    setupPoll();
#endif /* ifdef USE_LPDSP32_POLL */

    return 1;
}

/* ----------------------------------------------------------------------------
 * Function     : startProcessing
 * ----------------------------------------------------------------------------
 * Description  : Initiates a codec operation with the data in the current
 *                output buffer and sets the buffer status to PROCESSING
 * Inputs       : codec : an opaque handle to a codec object which process
 *                the data
 * Outputs      : Always returns 1
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static uint32_t startProcessing(CODEC codec)
{
    /* we can only get in here when we have a full buffer and the
     * DSP is ready to accept new data, so kick it
     */
    cs.status[cs.outBuffer] = PROCESSING;
    codecSetInputBuffer(codec, Buffer.input[cs.outBuffer],
                        cs.inLength[cs.outBuffer]);
    codecSetParameters(codec, &cs.parameters[cs.outBuffer]);
    if ((cs.parameters[cs.outBuffer].action & ENCODE) == ENCODE)
    {
        codecEncode(codec);
    }
    else
    {
        scratchLog(0xDEC0DE);
        scratchLog(cs.outBuffer);
        codecDecode(codec);
    }

#ifdef USE_LPDSP32_POLL
    setupPoll();
#endif /* ifdef USE_LPDSP32_POLL */

    return 1;
}

/* ----------------------------------------------------------------------------
 * Function     : unloadData
 * ----------------------------------------------------------------------------
 * Description  : fetches the output data from a codec operation and sets
 *                the current input buffer to EMPTY, allowing it to be
 *                recycled
 * Inputs       : codec : an opaque handle to a codec object which will
 *                provide the output data
 * Outputs      : Always returns 1
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static uint32_t unloadData(CODEC codec)
{
    /* send the data back to the ioProvider */
    ioWrite(cs.ioProvider, Buffer.output, codecGetOutputBufferUsed(codec));

    /* having dealt with the buffer we can discard it and switch */
    cs.status[cs.outBuffer] = EMPTY;
    cs.outBuffer = 1 - cs.outBuffer;

    return 1;
}

/* ----------------------------------------------------------------------------
 * Function     : processingComplete
 * ----------------------------------------------------------------------------
 * Description  : This will be invoked when we have finished processing by
 *                the codec the only thing allowed here is to transition the
 *                output status from processing to ready, anything else is
 *                an error
 * Inputs       : None
 * Outputs      : None
 * Assumptions  : This is expected to be invoked by the interrupt handler or
 *                if we are running in polling mode by the poll operation.
 * ------------------------------------------------------------------------- */
static void processingComplete(void)
{
    scratchLog(0xC0DED0 | cs.outBuffer);
    if (cs.status[cs.outBuffer] == PROCESSING)
    {
        cs.status[cs.outBuffer] = READY;
    }
    else
    {
        logFatal(0xDED);
    }
}

/* ----------------------------------------------------------------------------
 * Function     : readyToProcess
 * ----------------------------------------------------------------------------
 * Description  : Sets the configured flag
 * Inputs       : None
 * Outputs      : None
 * Assumptions  : This is expected to be invoked by the interrupt handler or
 *                if we are running in polling mode by the poll operation.
 * ------------------------------------------------------------------------- */
static void readyToProcess(void)
{
    cs.isConfigured = IS_CONFIGURED;
}

/* ----------------------------------------------------------------------------
 * Function     : audioProcessorInitialise
 * ----------------------------------------------------------------------------
 * Description  : Initialises the audio processor for use
 * Inputs       : codec : an opaque handle to a codec structure that will
 *                perform the encoding and decoding operations
 *                ioProvider : a pointer to an ioProvider that will be used
 *                as a source of and sink for data to be passed through the
 *                codec.
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
void audioProcessorInitialise(CODEC codec, pIOProvider ioProvider)
{
    codecSetNotifier(codec, &processingComplete);
    cs.isConfigured = NOT_CONFIGURED;
    cs.ioProvider = ioProvider;
    cs.status[0] = EMPTY;
    cs.status[1] = EMPTY;
    cs.inBuffer = 0;
    cs.outBuffer = 0;

#ifdef USE_LPDSP32_POLL
    pollPending = 0;
#endif /* ifdef USE_LPDSP32_POLL */
}

/* ----------------------------------------------------------------------------
 * Function     : audioProcess
 * ----------------------------------------------------------------------------
 * Description  : The main processing core of the object, this handles the
 *                state transitions, managing sample data being read from
 *                the ioProvider and results being sent back to it.
 * Inputs       : codec : an opaque handle to a codec structure that will
 *                perform the encoding and decoding operations
 * Outputs      : returns the return code from each of the state methods or
 *                1 if none of them have been used
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
uint32_t audioProcess(CODEC codec)
{
    if (cs.status[cs.inBuffer] == EMPTY)
    {
        return loadData();
    }

    if (cs.isConfigured == NOT_CONFIGURED)
    {
        return configure(codec);
    }
    else if (cs.isConfigured == IS_CONFIGURED)
    {
        if ((cs.status[cs.outBuffer] == FULL)
            && (cs.status[1 - cs.outBuffer] != PROCESSING))
        {
            return startProcessing(codec);
        }

        if (cs.status[cs.outBuffer] == READY)
        {
            return unloadData(codec);
        }
    }

#ifdef USE_LPDSP32_POLL
    return pollHandler(codec);
#else  /* ifdef USE_LPDSP32_POLL */
    return 1;
#endif /* ifdef USE_LPDSP32_POLL */
}

#ifndef USE_LPDSP32_POLL

/* ----------------------------------------------------------------------------
 * Function     : DSP0_IRQHandler
 * ----------------------------------------------------------------------------
 * Description  : Interrupt service handler which is activated when the
 *                LPDSP32 has completed it processing of the data
 * Inputs       : None
 * Outputs      : No specific outputs however the appropriate processing
 *                function will be invoked.
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
void DSP0_IRQHandler(void)
{
    scratchLog(0x111111);
    logEnd();
    if (cs.isConfigured == IS_CONFIGURED)
    {
        processingComplete();
    }
    else
    {
        readyToProcess();
    }
}

#endif /* ifndef USE_LPDSP32_POLL */
