# ----------------------------------------------------------------------------
# Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
# ON Semiconductor), All Rights Reserved
#
# This code is the property of ON Semiconductor and may not be redistributed
# in any form without prior written permission from ON Semiconductor.
# The terms of use and warranty for this code are covered by contractual
# agreements between ON Semiconductor and the licensee.
#
# This is Reusable Code.
#
# -----------------------------------------------------------------------------
#
from os import path
from Process import Process
from time import sleep, time
from random import randint
from struct import unpack

def bpLocation(line, result):
    '''
    Helper method which will parse output looking for a breakpoint being hit
    '''
    fields = line.rstrip().split(" ")
    if len(fields) > 3:
        if fields[0].startswith("Breakpoint"):
            if fields[1].endswith(","):
                return fields[2]
    return result
        
def intValue(line, result):
    '''
    Helper method which will parse an intValue from a given line
    '''
    fields = line.rstrip().split(" ")
    if len(fields) == 3:
        if fields[0].startswith("$"):
            if fields[1].endswith("="):
                return int(fields[2])
    return result

def dataParser(line, result):
    '''
    helper method to format data into an array of bytes
    '''
    fields = line.rstrip().split(":")
    if len(fields) == 2:
        if fields[0].startswith("0x"):
            myBytes = bytearray() if result is None else result
            myBytes.extend(bytearray(int(x) for x in fields[1].split("\t") if x))
            return myBytes
    return result

class gdbDriver(object):
    '''
    The gdbDriver object provides a mechanism to programmatically control GDB
    toallow us to run specific codecs under test conditions.
    '''

    def __init__(self, gdb, jlink, exe, jlinkId = None, stereo = False):
        '''
        Constructor, takes a number of fields as specified in the driver.py
        '''
        
        # Three flags control how noisy the logging is while running
        self.silentGDB = False
        self.noisyGDB = False
        self.noisySRV = False
        
        # set up the paths to GDB, and GDB Server as well as the jlink id
        self.gdb = path.realpath(gdb)
        self.jlink = path.realpath(jlink)
        self.exe = path.realpath(exe).replace("\\", "\\\\")
        self.output = None
        self.jlinkId = jlinkId

        # use standard GDBServer ports
        self.localHost = True
        self.GDBPort = '2331'
        self.SWOPort = '2332'
        self.IOPort  = '2333'
        
        # lots of stuf to set the initial conditions of GDB & Server        
        self.verify = True
        self.initRegisters = True
        self.silent = False
        self.singleRun = True        
        self.connectionTimeout = 0
        
        self.jlinkHost = "USB" if self.jlinkId is None else "USB=%s" % self.jlinkId
        self.jlinkScript = None
        self.jlinkSettings = None
        
        # Set up our device with some default communications settings
        self.targetDevice = "RSL10"
        self.targetInterface = "SWD"
        self.targetSpeed = "1000kHz"
        self.targetEndian = "little"
        
        # setup some default values for the codec parameters
        self.inPacketSize = 256;
        self.outPacketSize = 256;
        
        self.stereo = stereo
        self.channel = 0

        self.encode = True
        self.sampleRate = 16000
        self.frameSize = 160
        self.blockSize = 4
        self.mode = 1
        self.packed = False
        
        self.precision = 0
        self.maxPackets = 0
        self.packetCount = 0

        # we will maintain a history for debug purposes        
        self.history = []
        
        # and define the default breakpoints we will use
        self.inputBreakpoint = "ioRead"
        self.outputBreakpoint = "ioWrite"
        
    def __repr__(self):
        ''' Representation string for debug messages '''
        return "gdb: %s\njlink: %s\nexe: %s\n" % (self.gdb, self.jlink, self.exe)
    
    def enableOutput(self, outfile):
        self.output = outfile
    
    def _startDebugServer(self):
        '''
        Method to start the GDB server, this is being started as a process
        that can be read and written to
        '''
        command = [self.jlink]
        command.append('-select')
        command.append(self.jlinkHost)
        command.append('-device')
        command.append(self.targetDevice)
        command.append('-if')
        command.append(self.targetInterface)
        command.append('-speed')
        command.append(self.targetSpeed)
        command.append('-port')
        command.append(self.GDBPort)
        command.append('-noir')
        command.append('-singlerun')

        self.gdbServer = Process(command)
        self.gdbServer.start()
        
        sleep(0.5)
        while self.gdbServer.canRead():
            if self.noisySRV:
                print "<SRV> %s" % self.gdbServer.readline()
            else:
                self.gdbServer.readline();

    def _startDebugger(self):
        '''
        Start GDB, again this is being created as a process we can read and
        write
        '''
        command = [self.gdb]
        
        self.debugger = Process(command)
        self.debugger.start()        
        
        sleep(0.5)
        while self.debugger.canRead():
            if self.noisyGDB:
                print "<GDB> %s" % self.debugger.readline()
            else:
                self.debugger.readline()

    def _flushInput(self, parser = None):
        '''
        flush input method, ensures that we have no pending input coming
        from the debugger or server
        '''
        result = None
        
        while self.debugger.isAlive():
            if (self.debugger.canRead()):
                line = self.debugger.readline()
                if '___GDB___' in line:
                    break
                if parser is not None:
                    result = parser(line, result)
                if self.noisyGDB:
                    print "<GDB> %s" % line
            else:
                sleep(0.1)
        while (self.gdbServer.canRead()):
            if self.noisySRV:
                print "<SRV> %s" % self.gdbServer.readline()
            else:
                self.gdbServer.readline();
            
        return result
        
    def _execute(self, command, parser = None):
        '''
        Helper method to execute a single command in GDB, this takes an optional
        parser parameter that will be used to parse the response into various
        forms
        '''
        if not self.silentGDB:
            print "$ %s" % command
        self.debugger.send(command)
        return self._flushInput(parser)
        
    def _connectDebuggerAndServer(self):
        '''
        Some setup we need to do to connect the Debugger and Server
        '''
        # We set the prompt to something we can recognise, this is used when
        # parsing the output 
        self._execute("set prompt ___GDB___\\n")
        self._execute("target remote localhost:%s" % self.GDBPort)
    
    def _runToMain(self):
        '''
        reset the device, load our executable and run the program to main
        '''
        self._execute("monitor reset")
        self._execute("file '%s'" % self.exe)
        self._execute("load")
        self._execute("break main")
        self._execute("continue")
    
    def connect(self):
        '''
        Connect to the device using the debugger and debug server
        '''
        # start the debugger and debug server, we do it this way as we are ensuring that output
        # from the debug server appears after output from the gdb session for each command.
        self._startDebugger()
        self._startDebugServer()
        
        # so in theory at this stage we have a debug server running, connected to the device,
        # and a debugger to which we can send commands, let's try to connect the two
        self._connectDebuggerAndServer()
        
        # and load our program and run to main, that's us basically up and running
        self._runToMain()
    
    def disconnect(self):
        ''' Disconnect from the device by stopping the debugger '''
        self._execute("detach")
        self._execute("quit")
        if (self.debugger.isAlive()):
            self.debugger.stop()
    
    def _fetch(self, source, size):
        ''' read data from the source '''
        incoming = source.read(size)
        if len(incoming) != size:
            source.close()
        return incoming
    
    def _loadInputBuffer(self, infile):
        '''
        Load an input buffer from the input file and write it to the required
        location on the device for use by the application.
        '''
        if (self.maxPackets > 0):
            if (self.packetCount >= self.maxPackets):
                infile.close()  
        self.packetCount += 1
                
        # we are stopped on the ioRead function, so we need to set up the data inData
        # and size before continuing
        if infile.closed:
            return
        
        # little bit of messing around if we're testing a stereo dataset, this
        # will run the same data twice on different channels 
        if self.stereo == False or self.channel == 0:
            self._inData = self._fetch(infile, self.inPacketSize)
            
        if len(self._inData) == self.inPacketSize:        
            self._execute("set gdbParameters.action = %d" % self.makeAction(self._firstBlock))
            self._execute("set gdbParameters.modeAndChannel = %d" % (self.mode + self.channel))
            self._execute("set gdbInputSize = %d" % len(self._inData))
            dataBytes = ", ".join("0x{:02X}".format(c) for c in self._inData)
            self._execute("set gdbInput = { %s }" % dataBytes)
            self.history.append(self._inData)

            # more stereo jiggery pokery
            if self.stereo == True:
                self.channel = 0x10 - self.channel
                
            if self.stereo == False or self.channel == 0:
                self._firstBlock = False
            
    def dump(self, data):
        ''' Dump a data set to the console '''
        parts = [data[i:i+16] for i in range(0, len(data), 16)]
        for address in range(len(parts)):
            print "%s : %s" % ("{:04x}".format(address * 16), " ".join("{:02x}".format(c) for c in parts[address]))
        print ''
                            
    def dumpWords(self, data):
        ''' Dump a data set to the console '''
        parts = [data[i:i+16] for i in range(0, len(data), 16)]
        for address in range(len(parts)):
            line = "%s : " % "{:04x}".format(address * 16)
            words = [parts[address][i:i+4] for i in range(0, len(parts[address]), 4)]
            for word in words:
                line += "{:08x} ".format(unpack("<L", word)[0])
            print line
        print ''
                            
    def nextValues(self, expected, actual, size):
        index = 0;
        while index < len(expected):
            left = 0
            right = 0 
            for subindex in reversed(range(size)):
                left  = (left << 8) + expected[index + subindex]
                right = (right << 8) + actual[index + subindex]
            if left > 0x7FFF:
                left -= 0x10000
            if right > 0x7FFF:
                right -= 0x10000
            index += size
            yield (index, left, right)
        
    def findFails(self, expected, actual, size):
        ''' compare the expected and actual data '''
        differences = ''
        for (index, left, right) in self.nextValues(expected, actual, size):
            if abs(left - right) > self.precision:
                differences += "%s : %s %s\n" % ("{:04x}".format(index), "{:02x}".format(left), "{:02x}".format(right))
        return differences
              
    def capture(self, actual):
        if self.output is None:
            return
        self.output.write(actual)
          
    def _readOutputBuffer(self, goldfile):
        '''
        Read the output buffer from the device and push it through the
        comparison for failures, comparing against the golden data
        '''
        actualSize = self._execute("p gdbOutputSize", intValue)
        if actualSize != self.outPacketSize:
            raise ValueError("Incorrect response length %d, expected %d" % (actualSize, self.outPacketSize))

        self._execute("x/%dbx gdbOutput" % actualSize)
    
        dataBytes = self._execute("x/%dbu gdbOutput" % actualSize, dataParser)
        self.capture(dataBytes)
            
        if goldfile is None or goldfile.closed:
            return
        
        # more stereo jiggery pokery
        if self.stereo == False or self.channel == 0:
            self._outData = bytearray(self._fetch(goldfile, self.outPacketSize))
            
        if len(self._outData) == self.outPacketSize:
            
            if len(self._outData) != len(dataBytes):
                print "Unexpected data length returned"
                self.dump(self._outData);
                self.dump(dataBytes);
                raise ValueError("Incorrect response data length")
            
            fails = self.findFails(self._outData, dataBytes, 1 if self.encode else 2)
            if len(fails) > 0:
                print "Unexpected data returned\n%s" % fails
                self.dump(self.history[0])
                self.dump(self._outData);
                self.dump(dataBytes);
                raise ValueError("Incorrect response data")
                
            if (len(self.history) > 2):
                self.history.pop(0)

        dataBytes = ", ".join("0x00".format(c) for c in dataBytes)
        self._execute("set gdbOutput = { %s }" % dataBytes)

    def _bytesInFile(self, inFile):
        ''' calculate the file size '''
        inFile.seek(0, 2)
        result = inFile.tell()
        inFile.seek(0, 0)
        return result
    
    def loseFrame(self):
        lost = randint(1, 100) <= 0
        if lost:
            print "Lost Frame"
        return lost
    
    def makeAction(self, reset):
        '''
        Helper function to work out the next action depending on the encode
        or decode state
        '''
        action = 1 if self.encode else 2
        if reset == True:
            action += 4 if self.encode else 8
        action += 0x10 if self.packed else 0
        action += 0x40 if self.loseFrame() else 0
        return action
        
    def compare(self, infile, goldfile):
        '''
        The compare function is the main interface to the gdbDriver and this
        manages the reading and writing of the data files and marshalling the
        data through GDB to the device.
        '''
        
        # we know that we have two routines, ioRead and ioWrite that we will use to give us gaps
        # where we can insert or extract information, on entry to this call we expect to be stopped
        # on main
        
        # configure the codec with the given parameters
        self._firstBlock = True
        self._execute("set gdbParameters.sampleRate = %d" % self.sampleRate)
        self._execute("set gdbParameters.frameSize = %d" % self.frameSize)
        self._execute("set gdbParameters.blockSize = %d" % self.blockSize)
        
        # set the breakpoints we need
        self._execute("delete")
        self._execute("break gdbRead")
        self._execute("break gdbWriteComplete")
        
        startTime = time()
        bytesToProcess = self._bytesInFile(infile)
        if goldfile is not None:
            bytesToProcess += self._bytesInFile(goldfile)
        
        try:
            
            while not infile.closed:

                if not self.silentGDB:
                    self.dumpWords(self._execute("x/128bu &Buffer.configuration[0]", dataParser))
                    self.dumpWords(self._execute("x/256bu &Buffer.scratch[0]", dataParser))
                    self.dumpWords(self._execute("x/16bu &gdbParameters", dataParser))

                location = self._execute("continue", bpLocation)
            
                if (location == "gdbRead"):
                    if not self.silentGDB:
                        print "@gdbRead"
                    self._loadInputBuffer(infile)
            
                if (location == "gdbWriteComplete"):
                    if not self.silentGDB:
                        print "@gdbWriteComplete"
                    self._readOutputBuffer(goldfile)
                
                if not infile.closed:
                    bytesProcessed = infile.tell()
                    if goldfile is not None:
                        bytesProcessed += goldfile.tell()
                        
                    if bytesProcessed > 0:
                        timeToGo = ((time() - startTime) * (bytesToProcess - bytesProcessed)) / bytesProcessed
                        print "%d / %d - %d" % (bytesProcessed, bytesToProcess, timeToGo)
                    
        except ValueError as e:
            infile.close()
            if goldfile is not None:
                goldfile.close()
            raise e
