# ----------------------------------------------------------------------------
# Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
# ON Semiconductor), All Rights Reserved
#
# This code is the property of ON Semiconductor and may not be redistributed
# in any form without prior written permission from ON Semiconductor.
# The terms of use and warranty for this code are covered by contractual
# agreements between ON Semiconductor and the licensee.
#
# This is Reusable Code.
#
# -----------------------------------------------------------------------------
#

import json

class DSPSymbolIntegrator(object):
    '''
    Class to demonstrate how to retrieve symbol information from an LPDSP32
    ELF file
    '''

    def __init__(self, elf, output, tag, symbols):
        '''
        Constructor
        
        elf
            An ELFFile object that has already parsed the original file into
            parsable structures
        output
            The folder to which the generated files will be written
        tag
            A tag that is used when creating file and structure names so that
            they are identifiable by the CM3
        symbols
            JSON formatted file holding the symbols to be reported
        '''
        self._elf = elf
        self._output = output
        self._tag = tag
        self._symbols = symbols
        self._dwarf = elf.get_dwarf_info() if symbols and elf.has_dwarf_info() else None
        
    def _findDieByName(self, filename):
        '''
        Find the requested Debugging Information Entry (DIE) given the filename
        '''
        for CU in self._dwarf.iter_CUs():
            topDIE = CU.get_top_DIE()
            cuPath = topDIE.get_full_path()
            cuPath = cuPath.replace('\\', '/')
            cuPath = cuPath.split('/')
            if cuPath[-1] == filename:
                return topDIE
        raise ValueError("Unrecognised filename : %s", filename)

    def _findFunctionDieByName(self, parent, function):
        '''
        Given a DIE (Debugging Information Entry), locate the specified function 
        '''
        for child in parent.iter_children():
            if child.tag == 'DW_TAG_subprogram':
                fields = child.attributes['DW_AT_name'].value.split(' ')
                if function == fields[0]:
                    return child
        raise ValueError("Unrecognised function : %s", function)
                
    def _locateSymbols(self):
        '''
        Locate each symbol in our list of required symbols by looking up the
        debug information for the required source file and extracting each
        symbol in turn
        '''
        for source in self._requiredSymbols['sources']:
            die = self._findDieByName(source['file'])
            for function in source['functions']:
                functionDie = self._findFunctionDieByName(die, function['function'])
                yield (function['function'], functionDie.attributes['DW_AT_low_pc'].value, functionDie.attributes['DW_AT_high_pc'].value)
    
    def _createSymbolHeader(self, symbolGenerator):
        '''
        Creates a symbol header file with an entry for the start and end points
        of each of the required symbols.
        '''
        header = "%s/%s_symbols.h" % (self._output, self._tag)
        define = "_%s_DSP_SYMBOLS_H_" % (self._tag)
        with open(header, "w") as output:
            output.write("#ifndef %s\n" % define.upper())
            output.write("#define %s 1\n" % define.upper())
            output.write("\n")
            output.write("/* Auto generated file\n")
            output.write(" */\n")
            output.write("\n")
            for (symbol, start, end) in symbolGenerator():
                output.write("#define %s_dsp_%s_base    (0x%08x)\n" % (self._tag, symbol, start))
                output.write("#define %s_dsp_%s_end     (0x%08x)\n" % (self._tag, symbol, end))
            output.write("\n")
            output.write("#endif\n")
    
    def parseSymbols(self):
        '''
        This is the only public method in this class.
        
        If we have DWARF debug information available, load the JSON file to
        get the list of symbols and then publish a header file containing the
        relevant information  
        '''
        if self._dwarf == None:
            return

        with open(self._symbols, "r") as jsonFile:
            self._requiredSymbols = json.load(jsonFile)
        self._createSymbolHeader(self._locateSymbols)    
        
        